/*
 * Panasync Tools
 *
 * Copyright (C) 2001 Nuno Lopes
 * Copyright (C) 2000-2001 Carlos Baquero, Victor Fonte,
 * Paulo Sergio Almeida, Universidade do Minho
 *
 * panasync@gsd.di.uminho.pt - http://gsd.di.uminho.pt/panasync
 *
 * Departamento de Informatica, Universidade do Minho
 * Campus de Gualtar, 4710-057 Braga, Portugal
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 *
 */


#include "panacmds.h"

const int buffer_size = 1024;


 /** definition of debug level available for all modules/classes
	* 0 - no verbose
	* 1 - verbose command errors only
	* 2 - verbose errors for commands and pds
	* 3 - verbose errors for commands, pds, system (not implemented yet)
	*/
int PANA_debug_level = 1;


/** consider potencial warnings as errors and abort operation
 * 0 - don't
 * 1 - consider
 */
int PANA_warning_to_error = 0;



/** Initialize a base for use in panasync files.
 *
 * @param kind Char identifying kind of base.
 *  Possible values are 'u','s','r'.
 * @param id Unique identification for the base kind.
 * @param basefile String (\0 terminated) for base path.
 *  If null consider default base path that is /usr/share/panasync/base.
 * @return Pana error code. 0 if OK, code of error otherwise.
 */
int panainit(const char kind, const char * id, const char * basefile)
{
	// temporary buffer
	//
	char buffer[buffer_size];

	// base file name
	//
	string basename;

	// if argument basefile is null use default value for base path
	//
	if ( basefile != NULL)
	{
		basename = basefile;
	}
	else
	{
		basename = PTDEF_BASE_PATH;
	}

	// check base directory
	//
	string basedir = System::getDirName( basename.c_str(), buffer, buffer_size);
	
	// base directory is valid
	//
	if ( ! System::isDir( basedir.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("panainit: base directory is not valid."))
		return 1;
	}

	// base directory must have read-write permission
	//
	if ( ! System::isReadWriteAble( basedir.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("panainit: base directory does not have read/write access."))
		return 1;
	}

	// if base file exists...
	//
	if ( System::exists( basename.c_str()))
	{
		// if warn to error
		//
		if (PANA_warning_to_error)
		{
			ERROR( PANA_debug_commands,
				gettext("panainit: base file already exists."))
			return 1;
		}
		else // just warn about it...
		{
			WARN( PANA_debug_commands,
				"panainit: base file already exists, overwriting.")
		}

		// base file must be a regular file
		//
		if ( ! System::isFile( basename.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panainit: base file is not regular file."))
			return 1;
		}

		// base file must be writable
		//
		if ( ! System::isWritable( basename.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panainit: base file is not writable."))
			return 1;
		}
	} // end 'if base exists'
	
	// create a new base object
	//
	pts::base base;

	char str[32];
			
	// check kind for new base
	//
	switch (kind)
	{
		// unique kind of base
		//
		case 'u':
		case 'U':
			ERROR( PANA_debug_commands,
				gettext("panainit: [internal error] unique base not implemented yet."))
			return 1;
			break;

		// string kind of base
		//
		case 's':
		case 'S':
			// check if id field from arguments is valid
			//
			if ( id == NULL)
			{
				ERROR( PANA_debug_commands,
					gettext("panainit: ID field not valid for creation of string base."))
				return 1;
			}
			// update base object with values
			//
			base.kind = "string";
			base.id = id;
			base.counter = "1";
			break;

		// random kind of base
		//
		case 'r':
		case 'R':
			// use a random generated string for id field
			// counter should be '1' by convention to be able to distinguish
			//  from were it came {panainit,pananew}
			//
			base.kind = "random";
			base.id = System::generateString( str, 32);
			base.counter = "1";
			break;

		default:
			// kind argument not specified
			//
			if (PANA_warning_to_error)
			{
				ERROR( PANA_debug_commands,
					gettext("panainit: kind type not defined."))
				return 1;
			}
			else // just warn about it...
			{
				WARN( PANA_debug_commands,
				"panainit: using random kind for base (default).")
			}

			base.kind = "random";
			base.id = System::generateString( str, 32);
			base.counter = "1";
	}

	// write the base to file
	//
	base.write( basename.c_str());

	return 0;
}



/** Create a new PANASYNC file. 
 *
 * @param filename Path for the new file.
 * @param contentsfilename Path for optional file that will suply the
 *  new file contents. If null, left new file unchanged or create an
 *  empty file (if it does not exist already).
 * @param basefilename Path for the file containing the base info that
 *  will be used for this new file. If null, use default base path instead.
 * @return Pana error code. 0 if OK, error code otherwise.
 */
int pananew(const char * filename, const char * contentsfilename,
				const char * basefilename)
{
	// local variables
	//
	char buffer[buffer_size];
	pds pds;

	// checking pds directory
	//
	string directory = System::getDirName( filename, buffer, buffer_size);

	// directory section of filename is valid
	//
	if ( ! System::isDir( directory.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("pananew: target directory is not valid."))
		return 1;
	}

	// permissions on target directory must be read-write
	//
	if ( ! System::isReadWriteAble( directory.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("pananew: target directory does not have read/write access."))
		return 1;
	}
	
	// checking pds file
	//
	string pdsfile = directory;
	pdsfile += PTDEF_PDS_FILENAME;

	// if it exists already
	//
	if ( System::exists( pdsfile.c_str()))
	{
		// it must be a regular file 
		//
		if ( ! System::isFile( pdsfile.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("pananew: target pds file is not a regular file."))
			return 1;
		}

		// permissions on target pds file must be read-write
		//
		if ( ! System::isReadWriteAble( pdsfile.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("pananew: target pds file does not have read/write access."))
			return 1;
		}

		// if has pds file and warn to error flag...
		//
		if (PANA_warning_to_error)
		{
			// open pds directory
			//
			if ( ! pds.openDir( directory.c_str()))
			{
				ERROR( PANA_debug_commands,
					gettext("pananew: error in opening pds directory."))
				return 1;
			}

			// check if target is already a PANA file
			//
			System::getFileName( filename, buffer, buffer_size);
			
			if ( pds.hasFile( buffer ))
			{
				ERROR( PANA_debug_commands,
					gettext("pananew: new file is already a PANA file."))
				return 1;
			}

		} // end 'if has pds file and warn to error flag'
		
	} // end 'if pdsfile exists'

	// checking contents file
	//
	bool hascontents = (contentsfilename != 0);

	// if there is a contents file
	//
	if (hascontents)
	{
		// must have read access to the file
		//
		if ( ! System::isReadable(contentsfilename))
		{
			ERROR( PANA_debug_commands,
				gettext("pananew: contents file does not have read/write access."))
			return 1;
		}
	}

	// checking base file
	//
	string basefile;

	// if it was suplied a name, use it
	// 
	if ( basefilename != NULL)
	{
		basefile = basefilename;
	}
	// else use default value
	//
	else
	{
		basefile = PTDEF_BASE_PATH;
		WARN( PANA_debug_commands,
			"pananew: warning: using default base file path!" )
	}

	// base file must be a regular file
	//
	if ( ! System::isFile( basefile.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("pananew: base file is not regular file."))
		return 1;
	}

	// permissions on base file must be read-write
	//
	if ( ! System::isReadWriteAble( basefile.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("pananew: base file does not have read/write access."))
		return 1;
	}

	// checking target file
	//
	string targetname = System::getFileName( filename, buffer, buffer_size);

	// it must not be empty
	//
	if ( targetname.empty() )
	{
		ERROR( PANA_debug_commands,
			gettext("pananew: target file name is not valid (empty)."))
		return 1;
	}
	
	// target file exists ?
	//
	bool targetexist = System::exists(filename);

	// if it already exists
	//
	if ( targetexist )
	{
		// it must be a regular file 
		//
		if ( ! System::isFile( filename))
		{
			ERROR( PANA_debug_commands,
				gettext("pananew: target file is not a regular file."))
			return 1;
		}

		// target file must be readable
		//
		if ( ! System::isReadable( filename))
		{
			ERROR( PANA_debug_commands,
				gettext("pananew: target file does not have read access."))
			return 1;
		}

		// if there is a contents file verify more conditions...
		//
		if (hascontents)
		{
			// target file must have write access if there is a contents file
			//
			if ( ! System::isReadable( filename))
			{
				ERROR( PANA_debug_commands,
					gettext("pananew: target file does not have write access."))
				return 1;
			}
			
			// target file and contents file must not be the same
			//
			if ( ! System::isDifferent( filename, contentsfilename))
			{
				ERROR( PANA_debug_commands,
					gettext("pananew: target and contents files are the same."))
				return 1;
			}
		}
	} // end 'if target file exists'

	// operations
	
	// if has contents file,
	// copy from one to another
	// 
	if ( hascontents)
	{
		if ( !System::copy( contentsfilename, filename))
		{
			ERROR( PANA_debug_commands,
				gettext("pananew: error in copy from contents file to target file."))
			return 1;
		}
	}
	// if does not have contents and does not exist,
	// create a empty file
	//
	else if ( ! targetexist )
	{
		if ( !System::create( filename))
		{
			ERROR( PANA_debug_commands,
				gettext("pananew: error in creation of new file."))
			return 1;
		}
	}
	
	// open PDS directory
	//
	if ( ! pds.openDir( directory))
	{
		ERROR( PANA_debug_commands,
			gettext("pananew: error in opening pds directory."))
		return 1;
	}

	// read base info from file
	//
	pts::base base;
	base.read( basefile.c_str());

	// increment counter field from base
	//
	base.inc();

	// write base back to file
	//
	base.write( basefile.c_str());

	// check if file is already a PANASYNC file
	//
	if ( pds.hasFile(targetname.c_str()) )
	{
		// remove file
		//
		pds.removeFile( targetname.c_str());

		WARN(PANA_debug_commands,
			"pananew: new file already exists in pds directory, overwriting.")
	}
	
	// insert new file in pds directory with this base
	//
	if ( !pds.insertFile( targetname.c_str() , base))
	{
		ERROR( PANA_debug_commands,
			gettext("pananew: error in inserting file on pds directory."))
		return 1;
	}

	// close pds directory
	//
	if ( !pds.closeDir())
	{
		ERROR( PANA_debug_commands,
			gettext("pananew: error in closing pds directory."))
		return 1;
	}

	return 0;
}



/** Copies multiple sources to one target directory.
 *
 * It's a macro for panadirdup, which in turn is a macro for panadup.
 * It simply picks one source at a time and calls panadirdup with
 *  it and the target.
 *
 * @param args Number of strings stored in the array source.
 * @param source The array of strings (\0 terminated).
 * @param target The path of the target.
 * @return The result given by panadirdup.
 */
int panamultidup( int args, char ** source, char * target)
{
	int error =0;

	// check if target is directory
	//
	if ( ! System::isDir( target))
	{
		ERROR( PANA_debug_commands,
			gettext("panamultidup: target is not a directory."))
		return 1;
	}

	// for each source iterate
	// note: source can't be a directory (inspired on std copy command)
	//
	for (int i=0; i<args; i++)
	{
		// check if source is directory
		//
		if ( System::isDir(source[i]))
		{
			if (PANA_warning_to_error)
			{
				ERROR( PANA_debug_commands,
					gettext("panamultidup: source is directory."))
				return 1;
			}
			else // just warn about it...
			{
				WARN( PANA_debug_commands,
				"panamultidup: omitting source directory.")
			}
		}
		else // source is file ...
		{
			// call panadirdup
			//
			error = panadirdup( source[i], target);
	
			// if an error occurred
			//
			if ( error != 0)
				return error;
		}
	}

	return 0;
}



/** Copies source to target (directories).
 *
 * This function is a macro for panadup( sourcefile, targetfile).
 * It examines the arguments and calls panadup accordingly.
 *
 * @param source The path of the source file or directory.
 * @param target The path of the target file or directory.
 * @return PANA error code. 0 if OK, otherwise error code.
 */
int panadirdup( const char * source, const char * target)
{
	// local variables
	//
	int error;
	string targetname;
	string srcpdsname;
	pds pds;
	char buffer[buffer_size];

	// examine source type: file or directory

	// if source is file
	//
	if ( System::isFile(source))
	{
		// check if target is directory
		//
		if ( System::isDir( target))
		{
			// get source file name only and glue it to target
			//
			targetname = target;
			targetname += System::fileSystemSeparator;
			targetname += System::getFileName( source, buffer, buffer_size);

			return panadup( source, targetname.c_str());
		}

		// check if target is a file
		//
		if ( System::isFile( target))
		{
			// trivial case, call panadup directly
			//
			return panadup( source, target);
		}

		// check if target doesn't exist and has directory in path
		//
		System::getDirName( target, buffer, buffer_size);

		if ( !System::exists( target) && System::isDir(buffer))
		{
			// trivial case also, let panadup handle it
			//
			return panadup( source, target);
		}
		
		// target is not valid
		//
		ERROR( PANA_debug_commands,
			gettext("panadirdup: target is not valid."))
		return 1;

	}
	// elsif source is directory
	//
	else if ( System::isDir(source))
	{
		// check for read access to the source directory
		//
		if ( ! System::isReadable(source))
		{
			ERROR( PANA_debug_commands,
				gettext("panadirdup: source directory is not readable."))
			return 1;
		}

		// create source pds file name
		//
		srcpdsname = source;
		srcpdsname += System::fileSystemSeparator;
		srcpdsname += PTDEF_PDS_FILENAME;

		// check if source pds file exist
		//
		if ( ! System::exists(srcpdsname.c_str()))
		{
			// if warn to error flag
			//
			if (PANA_warning_to_error)
			{
				ERROR( PANA_debug_commands,
					gettext("panadirdup: source directory doesn't have any PANA file."))
				return 1;
			}
			else // just warn about it and continue...
			{
				WARN( PANA_debug_commands,
					"panadirdup: source directory doesn't have any PANA file, ignoring.")
				return 0;
			}

		} // end 'if source pds file exist'

		// check for read access to the source pds file
		//
		if ( ! System::isReadable(srcpdsname.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panadirdup: source pds file is not readable."))
			return 1;
		}

		// check for target being a directory
		//
		if ( ! System::isDir(target))
		{
			ERROR( PANA_debug_commands,
				gettext("panadirdup: source is directory but target is not."))
			return 1;
		}

		// open pds directory
		//
		if ( ! pds.openDir(source))
		{
			ERROR( PANA_debug_commands,
				gettext("panadirdup: error in opening pds source directory."))
			return 1;
		}

		// list pana files in directory
		//
		vector<string> filenames = pds.listFileNames();

		// close pds directory
		// ignore error checking because of read-only operations
		//
		pds.closeDir();

		// for each file, panadup'it
		//
		for (vector<string>::iterator i = filenames.begin();
				i!=filenames.end(); i++)
		{
			// create complete target name
			//
			targetname = target;
			targetname += System::fileSystemSeparator;
			targetname += *i;

			// prepend directory path to the source name
			//
			i->insert( (unsigned)0, (unsigned)1, System::fileSystemSeparator);
			i->insert( 0, source);

			// panadup complete source path to complete target path
			//
			error = panadup( i->c_str(), targetname.c_str());
			
			// if error occurred, abort
			// don't print message because it will be printed by panadup...
			//
			if ( error != 0)
			{
				return error;
			}
		} // end 'for each file in source'

	}	// end 'if source is directory'
	else
	// error: source is nothing valid!
	//
	{
		ERROR( PANA_debug_commands,
			gettext("panadirdup: source is not valid."))
		return 1;
	}

	return 0;
}


/** Copies source file to target.
 *
 * By copy, in PANASYNC terms, is meant contents copy 
 * (like the 'copy' command) and forking the time-stamp.
 * This operation, forking the time-stamp, is necessary
 * for placing in 'time' the two files and be able to compare them.
 *
 * @param source The path to the source file.
 * @param target The path to the target file.
 * @return PANA error code. 0 if OK, otherwise an error code.
 */
int panadup( const char * source, const char * target)
{
	// local variables
	
	// path of source directory 
	char srcdirname[buffer_size];
	// source file name (without prefix=path)
	char srcfilename[buffer_size];

	// path of target directory 
	char trgdirname[buffer_size];
	// target file name (without prefix=path)
	char trgfilename[buffer_size];

	System::getDirName( source, srcdirname, buffer_size);
	System::getFileName( source, srcfilename, buffer_size);

	System::getDirName( target, trgdirname, buffer_size);
	System::getFileName( target, trgfilename, buffer_size);

	// path of pds file for source
	string srcpdsname = srcdirname;
	srcpdsname += PTDEF_PDS_FILENAME;

	// path of pds file for target
	string trgpdsname = trgdirname;
	trgpdsname += PTDEF_PDS_FILENAME;

	// pds temporary object
	pds pds;


	// checking conditions

	
	// SOURCE SECTION

	// source directory
	//
	if ( ! System::isDir( srcdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: source directory is not valid."))
		return 1;
	}

	// source directory must be readable
	//
	if ( ! System::isReadable( srcdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: source directory is not readable."))
		return 1;
	}

	// source pds file
	//
	if ( ! System::isFile( srcpdsname.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: source pds file is not a regular file."))
		return 1;
	}

	// source pds file must be read write
	//
	if ( ! System::isReadWriteAble( srcpdsname.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: source pds file doesn't have read write permission."))
		return 1;
	}

	// source file
	//
	if ( ! System::isFile( source))
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: source file is not a regular file."))
		return 1;
	}

	// source file must be readable
	//
	if ( ! System::isReadable( source))
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: source file doesn't have read permission."))
		return 1;
	}


	// TARGET SECTION

	// target directory
	//
	if ( ! System::isDir( trgdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: target directory is not valid."))
		return 1;
	}

	// target directory must be read write 
	//
	if ( ! System::isReadWriteAble( trgdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: target directory doesn't have read write permission."))
		return 1;
	}

	// check if target pds file exists
	//
	if ( System::exists( trgpdsname.c_str()))
	{
		// target pds file is regular file
		//
		if ( ! System::isFile( trgpdsname.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panadup: target pds file is not a regular file."))
			return 1;
		}
	
		// target pds file must be read write
		//
		if ( ! System::isReadWriteAble( trgpdsname.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panadup: target pds file doesn't have read write permission."))
			return 1;
		}
	} // end 'if target pds file exists'

	// check if target file exists
	//
	if ( System::exists( target))
	{
		// if warn to error flag
		// FIXME: should this be verified?
		/*
		if ( PANA_warning_to_error)
		{
			ERROR( PANA_debug_commands,
				gettext("panadup: target file already exists."))
			return 1;
		}
		*/

		// target file must be regular file
		//
		if ( ! System::isFile( target))
		{
			ERROR( PANA_debug_commands,
				gettext("panadup: target file is not a regular file."))
			return 1;
		}
	
		// target file must be read write able
		//
		if ( ! System::isReadWriteAble( target))
		{
			ERROR( PANA_debug_commands,
				gettext("panadup: target file doesn't have read write permission."))
			return 1;
		}
	} // end 'if target file exists'

	// source and target must be different files
	//
	if ( ! System::isDifferent( source,target))
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: source and target are the same file."))
		return 1;
	}

	// if source and target dirs are the same
	//
	if ( ! System::isDifferent( srcdirname, trgdirname) )
	{
		// open pds directory
		//
		if ( ! pds.openDir( srcdirname))
		{
			ERROR( PANA_debug_commands,
				gettext("panadup: error when opening pds directory."))
			return 1;
		}

		// source must be a PANA file
		//
		if ( ! pds.hasFile( srcfilename))
		{
			ERROR( PANA_debug_commands,
				gettext("panadup: source is not a PANA file."))
			return 1;
		}

		// if warn flag is on
		//
		if (PANA_warning_to_error)
		{
			// if target already is a PANA file
			//
			if ( pds.hasFile( trgfilename))
			{
				ERROR( PANA_debug_commands,
					gettext("panadup: target already exists (is a PANA file)."))
				return 1;
			}
		} // end 'if warn flag'

		// close pds directory
		// doesn't check for errors because no write operation is done, yet 
		//
		pds.closeDir();
	}
	else // if directories are different
	{
		// open pds source directory
		//
		if ( ! pds.openDir( srcdirname))
		{
			ERROR( PANA_debug_commands,
				gettext("panadup: error when opening source pds directory."))
			return 1;
		}

		// source must be a PANA file
		//
		if ( ! pds.hasFile( srcfilename))
		{
			// if warn to error flag
			//
			if (PANA_warning_to_error)
			{
				ERROR( PANA_debug_commands,
					gettext("panadup: source is not a PANA file."))
				return 1;
			}
			else // just ignore it and continue...
			{
				WARN( PANA_debug_commands,
					"panadup: source is not a PANA file, ignoring.")
				return 0;
			}
		}

		// close pds directory
		// doesn't check for errors because no write operation is done, yet 
		//
		pds.closeDir();

		// if warn flag is on
		//
		if (PANA_warning_to_error)
		{
			// open pds target directory
			//
			if ( ! pds.openDir( trgdirname))
			{
				ERROR( PANA_debug_commands,
					gettext("panadup: error when opening target pds directory."))
				return 1;
			}
			
			// if target already is a PANA file
			//
			if ( pds.hasFile( trgfilename))
			{
				ERROR( PANA_debug_commands,
					gettext("panadup: target already exists (is PANA file)."))
				return 1;
			}

			// close pds directory
			// doesn't check for errors because no write operation is done, yet 
			//
			pds.closeDir();
		} // end 'if warn flag'

	} // end 'if-else directories are different'


	// operations

	// coping source to target
	//
	if ( ! System::copy( source, target))
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: error when coping source file to target."))
		return 1;
	}

	// open pds source directory
	//
	if ( ! pds.openDir( srcdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: error when opening source pds directory."))
		return 1;
	}

	// update source file if it is the case
	//
	pds.updateFile( srcfilename);
	
	// fork source file and retrieve a new filedata for the target file
	//
	pds::filedata fd = pds.forkFile( srcfilename);

	// close source pds direcory
	//
	if ( ! pds.closeDir())
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: error when closing source pds directory."))
		return 1;
	}

	// open pds target directory
	//
	if ( ! pds.openDir( trgdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: error when opening target pds directory."))
		return 1;
	}

	// remove target file if exists
	//
	if ( pds.hasFile( trgfilename))
	{
		pds.removeFile( trgfilename);

		WARN(PANA_debug_commands,
			"panadup: target file already exists in pds directory, overwriting.")
	}

	// insert new file with the filedata created before
	//
	if ( ! pds.insertFile( trgfilename, fd))
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: error when inserting target file on pds directory."))
		return 1;
	}
		
	// close target pds direcory
	//
	if ( ! pds.closeDir())
	{
		ERROR( PANA_debug_commands,
			gettext("panadup: error when closing target pds directory."))
		return 1;
	}
	
	return 0;
}


/** Moves multiple sources to one target directory.
 *
 * It's a macro for panadirmove, which in turn is a macro for panamove.
 * It simply picks one source at a time and calls panadirmove with
 *  it and the target.
 *
 * @param args Number of strings stored in the array source.
 * @param source The array of strings (\0 terminated).
 * @param target The path of the target.
 * @return The code error given by panadirmove.(0 for OK)
 */
int panamultimove( int args, char ** source, char * target)
{
	int error =0;

	// check if target is directory
	// (if there are multiple files to move to...)
	//
	if ( (args>1) && (!System::isDir(target)) )
	{
		ERROR( PANA_debug_commands,
			gettext("panamultimove: target is not a directory."))
		return 1;
	}

	// for each source iterate
	//
	for (int i=0; i<args; i++)
	{
		// call panadirmove
		//
		error = panadirmove( source[i], target);

		// if an error occurred
		//
		if ( error != 0)
			return error;
	}

	return 0;
}



/** Moves source to target (support for directories).
 *
 * This function is a macro for panamove( sourcefile, targetfile).
 * 
 * It examines the arguments and make the move procedure accordingly.
 * For regular files it just moves them, for pana-enabled files it moves
 * them and the meta-data binded to. 
 *
 * For directories, if the target directory already exists a new
 * sub-directory is created in the target (directory) and the source 
 * contents is moved into it. If the target does not exist, the contents of
 * source is moved into the newly created target directory.
 *
 * @param source The path of the source file or directory.
 * @param target The path of the target file or directory.
 * @return PANA error code. 0 if OK, otherwise error code.
 */
int panadirmove( const char * source, const char * target)
{
	// local variables
	//
	int error;
	string targetname;
	string srcpdsname;
	pds pds;
	char buffer[buffer_size];

	// examine source type: file or directory

	// if source is file
	//
	if ( System::isFile(source))
	{
		// check if target is directory
		//
		if ( System::isDir( target))
		{
			// get source file name only and glue it to target
			//
			targetname = target;
			targetname += System::fileSystemSeparator;
			targetname += System::getFileName( source, buffer, buffer_size);

			return panamove( source, targetname.c_str());
		}

		// check if target is a file
		//
		if ( System::isFile( target))
		{
			// trivial case, call panamove directly
			//
			return panamove( source, target);
		}

		// check if target doesn't exist and has directory in path
		//
		System::getDirName( target, buffer, buffer_size);

		if ( !System::exists( target) && System::isDir(buffer))
		{
			// trivial case also, let panamove handle it
			//
			return panamove( source, target);
		}
		
		// target is not valid
		//
		ERROR( PANA_debug_commands,
			gettext("panadirmove: target is not valid."))
		return 1;

	}
	// elsif source is directory
	//
	else if ( System::isDir(source))
	{
		// local variable
		string targetbase;
		
		// check if target exists
		//
		if ( System::exists(target))
		{
			// check for target being a directory
			//
			if ( ! System::isDir(target))
			{
				ERROR( PANA_debug_commands,
					gettext("panadirmove: target is not directory."))
				return 1;
			}	
			
			// problem description: when moving a to b,
			// if b already exist a new sub-dir must be created b/a
			// if not, a/* should go into b/*
			
			// remove ending fileSystemSeparator from source path
			strncpy( buffer, source, buffer_size);
			System::removeLastSeparator( buffer);

			// get source sub-dir name
			//
			char bufferaux[buffer_size];
			System::getFileName( buffer, bufferaux, buffer_size);

			// set targetbase to target + buffer
			//
			targetbase = target;
			targetbase += System::fileSystemSeparator;
			targetbase += bufferaux;
			
			// make new sub-dir on target
			if ( ! System::createDir(targetbase.c_str()) )
			{
				ERROR( PANA_debug_commands,
					gettext("panadirmove: error on making target sub-directory."))
				return 1;
			}
		}	
		// target directory does not exist
		//
		else
		{
			if ( ! System::createDir(target) )
			{
				ERROR( PANA_debug_commands,
					gettext("panadirmove: error on making target directory."))
				return 1;
			}

			// set targetbase to target
			//
			targetbase = target;
		}

		// check for read-write access to the source directory
		//
		if ( ! System::isReadWriteAble(source))
		{
			ERROR( PANA_debug_commands,
				gettext("panadirmove: source directory is not read-write able."))
			return 1;
		}

		// check for read-write access to the targetbase directory
		//
		if ( ! System::isReadWriteAble( targetbase.c_str()) )
		{
			ERROR( PANA_debug_commands,
				gettext("panadirmove: target directory is not read-write able."))
			return 1;
		}

		// create source pds file name
		//
		srcpdsname = source;
		srcpdsname += System::fileSystemSeparator;
		srcpdsname += PTDEF_PDS_FILENAME;

		// check if source pds file exist and
		// for read access to the source pds file
		//
		if ( System::exists(srcpdsname.c_str())
			&& ! System::isReadable(srcpdsname.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panadirmove: source pds file is not readable."))
			return 1;
		}

		// open pds directory
		//
		if ( ! pds.openDir(source))
		{
			ERROR( PANA_debug_commands,
				gettext("panadirmove: error in opening pds source directory."))
			return 1;
		}

		// list pana files in directory
		//
		vector<string> filenames = pds.listFileNames();

		// close pds directory
		// ignore error checking because of read-only operations
		//
		pds.closeDir();

		// for each pana-file, panamove'it
		//
		for (vector<string>::iterator i = filenames.begin();
				i!=filenames.end(); i++)
		{
			// create complete target name
			//
			targetname = targetbase;
			targetname += System::fileSystemSeparator;
			targetname += *i;

			// prepend directory path to the source name
			//
			i->insert( (unsigned)0, (unsigned)1, System::fileSystemSeparator);
			i->insert( 0, source);

			// panamove complete source path to complete target path
			//
			error = panamove( i->c_str(), targetname.c_str());
			
			// if error occurred, abort
			// don't print message because it will be printed by panamove...
			//
			if ( error != 0)
			{
				return error;
			}
		} // end 'for each pana-file in source'

		// moving of regular files
		//
		filenames = System::listFiles(source);

		for (vector<string>::iterator i = filenames.begin();
				i!=filenames.end(); i++)
		{
			// create complete target name
			//
			targetname = targetbase;
			targetname += System::fileSystemSeparator;
			targetname += *i;

			// prepend directory path to the source name
			//
			i->insert( (unsigned)0, (unsigned)1, System::fileSystemSeparator);
			i->insert( 0, source);

			// move complete source path to complete target path
			//
			if( ! System::copy( i->c_str(), targetname.c_str()))
			{
				ERROR( PANA_debug_commands,
					gettext("panadirmove: error in coping source file to target directory."))
				return 1;
			}
			
			if( ! System::remove( i->c_str()))
			{
				ERROR( PANA_debug_commands,
					gettext("panadirmove: error in removing source file."))
				return 1;
			}
		} // end 'for each regular file in source'

		
		// moving of sub-directories
		//
		filenames = System::listDirs(source);

		for (vector<string>::iterator i = filenames.begin();
				i!=filenames.end(); i++)
		{
			// create complete target name
			//
			targetname = targetbase;
			targetname += System::fileSystemSeparator;
			targetname += *i;

			// prepend directory path to the source name
			//
			i->insert( (unsigned)0, (unsigned)1, System::fileSystemSeparator);
			i->insert( 0, source);

			// panadirmove complete source path to complete target path
			//
			error = panadirmove( i->c_str(), targetname.c_str());
			
			if ( error != 0)
				return error;
		
		} // end 'for each sub-directory in source'

		// remove source directory
		//
		if( ! System::removeDir(source))
		{
			ERROR( PANA_debug_commands,
				gettext("panadirmove: error in removing source directory."))
			return 1;
		}

	}	// end 'if source is directory'
	else
	// error: source is nothing valid!
	//
	{
		ERROR( PANA_debug_commands,
			gettext("panadirmove: source is not valid."))
		return 1;
	}

	return 0;
}




/** Move source file to target.
 *
 * Moving is creating a file, target, with the
 *  contents of the source and removing the source.
 * With a PANASYNC file, the time-stamp must also be 'moved':
 *  erased from the source pds file and inserted into the target pds
 *  directory
 *
 * @param source The path to the source file.
 * @param target The path to the target file.
 * @return PANA error code. 0 if OK, otherwise an error code.
 */
int panamove( const char * source, const char * target)
{
	// local variables
	
	// path of source directory 
	char srcdirname[buffer_size];
	// source file name (without prefix)
	char srcfilename[buffer_size];

	// path of target directory 
	char trgdirname[buffer_size];
	// target file name (without prefix)
	char trgfilename[buffer_size];

	System::getDirName( source, srcdirname, buffer_size);
	System::getFileName( source, srcfilename, buffer_size);

	System::getDirName( target, trgdirname, buffer_size);
	System::getFileName( target, trgfilename, buffer_size);

	// path of pds file for source
	string srcpdsname = srcdirname;
	srcpdsname += PTDEF_PDS_FILENAME;

	// path of pds file for target
	string trgpdsname = trgdirname;
	trgpdsname += PTDEF_PDS_FILENAME;

	// pds temporary object
	pds pds;


	// checking conditions

	// SOURCE SECTION

	// source directory
	//
	if ( ! System::isDir( srcdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: source directory is not valid."))
		return 1;
	}

	// source directory must be read write
	//
	if ( ! System::isReadWriteAble( srcdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: source directory doesn't have read write permission."))
		return 1;
	}

	// source pds file
	//
	if ( ! System::isFile( srcpdsname.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: source pds file is not a regular file."))
		return 1;
	}

	// source pds file must be read write
	//
	if ( ! System::isReadWriteAble( srcpdsname.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: source pds file doesn't have read write permission."))
		return 1;
	}

	// source file
	//
	if ( ! System::isFile( source))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: source file is not a regular file."))
		return 1;
	}

	// source file must be readable
	//
	if ( ! System::isReadable( source))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: source file doesn't have read permission."))
		return 1;
	}


	// TARGET SECTION

	// target directory
	//
	if ( ! System::isDir( trgdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: target directory is not valid."))
		return 1;
	}

	// target directory must be read write 
	//
	if ( ! System::isReadWriteAble( trgdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: target directory doesn't have read write permission."))
		return 1;
	}

	// check if target pds file exists
	//
	if ( System::exists( trgpdsname.c_str()))
	{
		// target pds file is regular file
		//
		if ( ! System::isFile( trgpdsname.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panamove: target pds file is not a regular file."))
			return 1;
		}
	
		// target pds file must be read write
		//
		if ( ! System::isReadWriteAble( trgpdsname.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panamove: target pds file doesn't have read write permission."))
			return 1;
		}
	} // end 'if target pds file exists'

	// check if target file exists
	//
	if ( System::exists( target))
	{
		// if warn to error flag
		// FIXME: should this be verified?
		/*
		if ( PANA_warning_to_error)
		{
			ERROR( PANA_debug_commands,
				gettext("panamove: target file already exists."))
			return 1;
		}
		*/

		// target file must be regular file
		//
		if ( ! System::isFile( target))
		{
			ERROR( PANA_debug_commands,
				gettext("panamove: target file is not a regular file."))
			return 1;
		}
	
		// target file must be read write able
		//
		if ( ! System::isReadWriteAble( target))
		{
			ERROR( PANA_debug_commands,
				gettext("panamove: target file doesn't have read write permission."))
			return 1;
		}
	} // end 'if target file exists'

	// source and target must be different files
	//
	if ( ! System::isDifferent( source,target))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: source and target are the same file."))
		return 1;
	}

	// if source and target dirs are the same
	//
	if ( ! System::isDifferent( srcdirname, trgdirname) )
	{
		// open pds directory
		//
		if ( ! pds.openDir( srcdirname))
		{
			ERROR( PANA_debug_commands,
				gettext("panamove: error when opening pds directory."))
			return 1;
		}

		// source must be a PANA file
		//
		if ( ! pds.hasFile( srcfilename))
		{
			ERROR( PANA_debug_commands,
				gettext("panamove: source is not a PANA file."))
			return 1;
		}

		// if warn flag is on
		//
		if (PANA_warning_to_error)
		{
			// if target already is a PANA file
			//
			if ( pds.hasFile( trgfilename))
			{
				ERROR( PANA_debug_commands,
					gettext("panamove: target already exists (is a PANA file)."))
				return 1;
			}
		} // end 'if warn flag'

		// close pds directory
		// doesn't check for errors because no write operation is done, yet 
		//
		pds.closeDir();
	}
	else // if directories are different
	{
		// open pds source directory
		//
		if ( ! pds.openDir( srcdirname))
		{
			ERROR( PANA_debug_commands,
				gettext("panamove: error when opening source pds directory."))
			return 1;
		}

		// source must be a PANA file
		//
		if ( ! pds.hasFile( srcfilename))
		{
			// if warn to error flag
			//
			if (PANA_warning_to_error)
			{
				ERROR( PANA_debug_commands,
					gettext("panamove: source is not a PANA file."))
				return 1;
			}
			else // just ignore it and continue...
			{
				WARN( PANA_debug_commands,
					"panamove: source is not a PANA file, ignoring.")
				return 0;
			}
		}

		// close pds directory
		// doesn't check for errors because no write operation is done, yet 
		//
		pds.closeDir();

		// if warn flag is on
		//
		if (PANA_warning_to_error)
		{
			// open pds target directory
			//
			if ( ! pds.openDir( trgdirname))
			{
				ERROR( PANA_debug_commands,
					gettext("panamove: error when opening target pds directory."))
				return 1;
			}
			
			// if target already is a PANA file
			//
			if ( pds.hasFile( trgfilename))
			{
				ERROR( PANA_debug_commands,
					gettext("panamove: target already exists (is PANA file)."))
				return 1;
			}

			// close pds directory
			// doesn't check for errors because no write operation is done, yet 
			//
			pds.closeDir();
		} // end 'if warn flag'

	} // end 'if-else directories are different'


	// operations

	// coping source to target
	//
	if ( ! System::copy( source, target))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: error when coping source file to target."))
		return 1;
	}

	// open pds source directory
	//
	if ( ! pds.openDir( srcdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: error when opening source pds directory."))
		return 1;
	}

	// get filedata structure for source
	//
	pds::filedata fd = pds.getFileData( srcfilename);

	// remove source file from pds file
	//
	if ( ! pds.removeFile( srcfilename))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: error when removing source from pds directory."))
		return 1;
	}

	// remove source from disk
	//
	if ( ! System::remove( source))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: error when removing source from disk."))
		return 1;
	}
	
	// close source pds direcory
	//
	if ( ! pds.closeDir())
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: error when closing source pds directory."))
		return 1;
	}

	// open pds target directory
	//
	if ( ! pds.openDir( trgdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: error when opening target pds directory."))
		return 1;
	}

	// remove target file if exists
	//
	if ( pds.hasFile( trgfilename))
	{
		pds.removeFile( trgfilename);

		WARN(PANA_debug_commands,
			"panamove: target file already exists in pds directory, overwriting.")
	}

	// insert new file with the filedata created before
	//
	if ( ! pds.insertFile( trgfilename, fd))
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: error when inserting target file on pds directory."))
		return 1;
	}
		
	// close target pds direcory
	//
	if ( ! pds.closeDir())
	{
		ERROR( PANA_debug_commands,
			gettext("panamove: error when closing target pds directory."))
		return 1;
	}
	
	return 0;
}



/** Joins two files, resulting just one of them, the target.
 *
 * Joining two files is removing the source, placing the newer (or older)
 * contents in the target file and updating the time-stamp.
 *
 * If the two files are concurrent versions and if a contents file is
 *  suplied, the target will join the two time-stamps and the contents
 *  will be copied from it.
 *
 * @param source The path to the source file.
 * @param target The path to the second source and target file.
 * @param contents The path to an alternative contents file
 *  (may not be needed).
 * @param older Flag for reversing decision and storing the older file
 *  instead of the new one. 1 for reversing otherwise 0.
 * @return PANA error code. 0 if OK, otherwise an error code.
 */
int panajoin( const char * target, const char * source,
				const char * contents, int older)
{
	// local variables
	//
	bool reverse = (older==1);

	// path of source directory 
	char srcdirname[buffer_size];
	// source file name (without prefix)
	char srcfilename[buffer_size];

	// path of target directory 
	char trgdirname[buffer_size];
	// target file name (without prefix)
	char trgfilename[buffer_size];

	System::getDirName( source, srcdirname, buffer_size);
	System::getFileName( source, srcfilename, buffer_size);

	System::getDirName( target, trgdirname, buffer_size);
	System::getFileName( target, trgfilename, buffer_size);

	// path of pds file for source
	string srcpdsname = srcdirname;
	srcpdsname += PTDEF_PDS_FILENAME;

	// path of pds file for target
	string trgpdsname = trgdirname;
	trgpdsname += PTDEF_PDS_FILENAME;

	// pds temporary object
	pds srcpds;
	pds trgpds;

	// filedata for source and target for comparison
	pds::filedata srcfd;
	pds::filedata trgfd;

	// pts temporary object
	pts pts;


	// checking conditions

	// SOURCE SECTION

	// source directory
	//
	if ( ! System::isDir( srcdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: source directory is not valid."))
		return 1;
	}

	// source directory must be read write
	//
	if ( ! System::isReadWriteAble( srcdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: source directory doesn't have read write permission."))
		return 1;
	}

	// source pds file
	//
	if ( ! System::isFile( srcpdsname.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: source pds file is not a regular file."))
		return 1;
	}

	// source pds file must be read write
	//
	if ( ! System::isReadWriteAble( srcpdsname.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: source pds file doesn't have read write permission."))
		return 1;
	}

	// source file
	//
	if ( ! System::isFile( source))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: source file is not a regular file."))
		return 1;
	}

	// source file must be readable
	//
	if ( ! System::isReadable( source))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: source file doesn't have read permission."))
		return 1;
	}


	// TARGET SECTION

	// target directory
	//
	if ( ! System::isDir( trgdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: target directory is not valid."))
		return 1;
	}

	// target directory must be read write 
	//
	if ( ! System::isReadWriteAble( trgdirname))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: target directory doesn't have read write permission."))
		return 1;
	}

	// target pds file is regular file
	//
	if ( ! System::isFile( trgpdsname.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: target pds file is not a regular file."))
		return 1;
	}
	
	// target pds file must be read write
	//
	if ( ! System::isReadWriteAble( trgpdsname.c_str()))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: target pds file doesn't have read write permission."))
		return 1;
	}

	// target file must be regular file
	//
	if ( ! System::isFile( target))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: target file is not a regular file."))
		return 1;
	}
	
	// target file must be read write able
	//
	if ( ! System::isReadWriteAble( target))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: target file doesn't have read write permission."))
		return 1;
	}

	// source and target must be different files
	//
	if ( ! System::isDifferent( source,target))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: source and target are the same file."))
		return 1;
	}

	// check if directories are the same
	//
	bool samedir = ! System::isDifferent( srcdirname, trgdirname);

	// if source and target dirs are the same
	//
	if ( samedir )
	{
		// open pds directory
		//
		if ( ! trgpds.openDir( srcdirname))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: error when opening pds directory."))
			return 1;
		}

		// source must be a PANA file
		//
		if ( ! trgpds.hasFile( srcfilename))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: source is not a PANA file."))
			return 1;
		}

		// target must be a PANA file
		//
		if ( ! trgpds.hasFile( trgfilename))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: target is not a PANA file."))
			return 1;
		}

		// source and target must be comparable
		// (it should not depend on file update's)
		//
		pts = trgpds.getFileData(srcfilename).stamp;

		if ( trgpds.getFileData(trgfilename).stamp.uncomparable(pts) )
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: source and target come from a different lineage (uncomparable)."))
			return 1;
		}

		// update hash and stamp for source and target files
		//
		if ( ! trgpds.updateFile( srcfilename))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: error when updating source file in pds directory."))
			return 1;
		}

		if ( ! trgpds.updateFile( trgfilename))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: error when updating target file in pds directory."))
			return 1;
		}

		// retrieve filedata strucuture for both files
		//
		srcfd = trgpds.getFileData( srcfilename);
		trgfd = trgpds.getFileData( trgfilename);
		
	}
	else // if directories are different
	{
		// open pds source directory
		//
		if ( ! srcpds.openDir( srcdirname))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: error when opening source pds directory."))
			return 1;
		}

		// source must be a PANA file
		//
		if ( ! srcpds.hasFile( srcfilename))
		{
				ERROR( PANA_debug_commands,
					gettext("panajoin: source is not a PANA file."))
				return 1;
		}

		// update hash and stamp for source file
		//
		if ( ! srcpds.updateFile( srcfilename))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: error when updating source file in pds directory."))
			return 1;
		}

		// retrieve source filedata for future use
		//
		srcfd = srcpds.getFileData( srcfilename);

		
		// open pds target directory
		//
		if ( ! trgpds.openDir( trgdirname))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: error when opening target pds directory."))
			return 1;
		}
			
		// target must be a PANA file
		//
		if ( ! trgpds.hasFile( trgfilename))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: target is not a PANA file."))
			return 1;
		}

		// source and target must be comparable
		// (must not depend on file update's)
		//
		pts = srcfd.stamp;

		if ( trgpds.getFileData(trgfilename).stamp.uncomparable(pts) )
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: source and target come from a different lineage (uncomparable)."))
			return 1;
		}

		// update hash and stamp for target file
		//
		if ( ! trgpds.updateFile( trgfilename))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: error when updating target file in pds directory."))
			return 1;
		}

		// retrieve target filedata for future use
		//
		trgfd = trgpds.getFileData( trgfilename);

	} // end 'if-else directories are different'

	// check contents file if any
	//
	if ( contents != NULL)
	{
		// check if contents is a file
		//
		if ( ! System::isFile( contents))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: contents file is not a regular file."))
			return 1;
		}

		// check if contents file is readable
		//
		if ( ! System::isReadable( contents))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: contents file doesn't have read permission."))
			return 1;
		}
	}


	// operations

	// compare time-stamps to evaluate what is newer or older
	
	// if time-stamps are equivalents
	//
	if ( srcfd.stamp.equivalent( trgfd.stamp))
	{
		WARN( PANA_debug_commands,
			"panajoin: files are equivalent.")
	}
	// if target is newer than source
	else if ( trgfd.stamp.dominates( srcfd.stamp))
	{
		// check if reverse flag is true
		//
		if ( reverse )
		{
			WARN( PANA_debug_commands,
				"panajoin: source is older than target.")	

			// coping source to target
			//
			if ( ! System::copy( source, target))
			{
				ERROR( PANA_debug_commands,
					gettext("panajoin: error when coping source file to target."))
				return 1;
			}
		}
		else
		{
			WARN( PANA_debug_commands,
				"panajoin: target is newer than source.")
		} // end 'if reverse flag'
	}
	// if source is newer that target
	else if (srcfd.stamp.dominates( trgfd.stamp))
	{
		// check if reverse flag is true
		//
		if ( reverse )
		{
			WARN( PANA_debug_commands,
				"panajoin: target is older than source.")
		}
		else
		{
			WARN( PANA_debug_commands,
				"panajoin: source is newer than target.")
			
			// coping source to target
			//
			if ( ! System::copy( source, target))
			{
				ERROR( PANA_debug_commands,
					gettext("panajoin: error when coping source file to target."))
				return 1;
			}
		} // end 'if reverse flag'
	}
	// if source and target are concurrent
	else 
	{
		// if there is a contents file
		//
		if ( contents != NULL)
		{
			WARN( PANA_debug_commands,
				"panajoin: source and target are concurrent, using contents file.")

			// copy contents file to target
			//
			if ( ! System::copy( contents, target))
			{
				ERROR( PANA_debug_commands,
					gettext("panajoin: error when coping contents file to target."))
				return 1;
			}
		}
		else // if theres no contents file
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: source and target are concurrent."))
			return 1;
		}

	} // end 'if are equivalent, newer, older, concurrent'

	
	// update information on pds files for target and source

	// remove target file from pds file
	//
	if ( ! trgpds.removeFile( trgfilename))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: error when removing target from pds directory."))
		return 1;
	}

	// join time-stamps
	// (order is indifferent)
	//
	pts = srcfd.stamp;
	pts.join( trgfd.stamp);

	// insert new file with the time-stamp created
	// implicit evaluation of md5 hash
	//
	if ( ! trgpds.insertFile( trgfilename, pts))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: error when inserting new target file on pds directory."))
		return 1;
	}

	// if directories are different
	//
	if ( ! samedir)
	{
		// remove source file from pds file
		//
		if ( ! srcpds.removeFile( srcfilename))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: error when removing source from pds directory."))
			return 1;
		}

		// close source pds direcory
		//
		if ( ! srcpds.closeDir())
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: error when closing source pds directory."))
			return 1;
		}
	}
	else // same directory for source and target
	{
		// remove source file from pds file
		//
		if ( ! trgpds.removeFile( srcfilename))
		{
			ERROR( PANA_debug_commands,
				gettext("panajoin: error when removing source from pds directory."))
			return 1;
		}
	}

	// close target pds direcory
	//
	if ( ! trgpds.closeDir())
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: error when closing target pds directory."))
		return 1;
	}

	// remove source from disk
	//
	if ( ! System::remove( source))
	{
		ERROR( PANA_debug_commands,
			gettext("panajoin: error when removing source from disk."))
		return 1;
	}

	return 0;
}



/** Process directory panasync contents by base */
int panalistgroup( const char * pathname, int options,
		map< string, vector<string> > & cluster )
{
	pds vpds;
	pds::filedata fd;
	
	// check if pathname is directory
	//
	if ( ! System::isDir( pathname))
	{
		ERROR( PANA_debug_commands,
			gettext("panalistgroup: path is not directory."))
		return 1;
	}

	// directory must be readable
	//
	if ( ! System::isReadable( pathname))
	{
		ERROR( PANA_debug_commands,
			gettext("panalist: directory doesn't have read permission."))
		return 1;
	}

	// check if pds file exists
	//
	string pdsfilename = pathname;
	pdsfilename += System::fileSystemSeparator;
	pdsfilename += PTDEF_PDS_FILENAME;
	
	if ( System::exists( pdsfilename.c_str() ))
	{
		// pds file is regular file
		//
		if ( ! System::isFile( pdsfilename.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panalistgroup: pds file is not a regular file."))
			return 1;
		}
	
		// pds file must be readable
		//
		if ( ! System::isReadable( pdsfilename.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panalistgroup: pds file doesn't have read permission."))
			return 1;
		}
	} // end 'if pds file exists'

	// open pds directory
	//
	if ( ! vpds.openDir( pathname))
	{
		ERROR( PANA_debug_commands,
			gettext("panalistgroup: error in opening pds directory."))
		return 1;
	}

	// get files
	//
	vector<string> l = vpds.listFileNames();
	
	// for each pana-file
	//
	for ( vector<string>::iterator i = l.begin(); i!=l.end(); i++)
	{
		// create complete path name for pana-file
		//
		string path = pathname;
		if ( path[ path.size()-1] != System::fileSystemSeparator)
			path += System::fileSystemSeparator;
		path += *i;

		// update information
		//
		vpds.updateFile( i->c_str() );

		// get filedata for file
		//
		fd = vpds.getFileData( i->c_str());

		string base = fd.stamp.getBase();

		// insert it into the map
		//
		cluster[ base ].push_back( path);

	}

	// process sub-dirs if option
	//
	if ( options & 8 )
	{
		l = System::listDirs( pathname);
		for ( vector<string>::iterator i=l.begin(); i!=l.end(); i++)
		{
			panalistgroup( i->c_str(), options, cluster);
		}
	}

	return 0;
}

/** Lists an array of paths.
 *
 * Is a macro for panalist.
 */
int panamultilist( int args, char ** path, int options, const char * compare)
{
	int error =0;

	// check if cluster option is enabled
	// FIXME: needs a clean-up
	//
	if ( options & 16)
	{
		map< string, vector<string> > cluster;
	
		for ( int k=0; k<args; k++)
		{
			panalistgroup( path[k], options, cluster);
		}

		// check if there's no paths and run with current directory
		if ( args == 0)
		{
			panalistgroup( System::currentDirectory, options, cluster);
		}
	
		// print information
		for ( map< string, vector<string> >::iterator i= cluster.begin();
					i!= cluster.end(); i++)
		{
			// print base:
			cout << i->first.c_str() << " :" << endl;
			
			// print every filename
			//
			for ( vector<string>::iterator j = i->second.begin();
					j!=i->second.end(); j++)
			{
//				cout << " " << j->c_str() << endl;
				panalist( j->c_str(), options, compare);
			}
			cout << endl;
		}
		return 0;
	} // end 'if cluster option'
	
	// if there's no paths
	//
	if ( args == 0)
	{
		return panalist( System::currentDirectory, options, compare);
	}

	// for each path iterate
	//
	for (int i=0; i<args; i++)
	{
		// call panadirmove
		//
		error = panalist( path[i], options, compare);

		// if an error occurred
		//
		if ( error != 0)
			return error;
	}

	return 0;
}


/** Lists panafiles in a directory.
 *
 * @param pathname Directory or panafile to show info of...
 * @param options Options suported
 *  P_list_rec : recursive listing of all sub-dirs,
 *  P_list_long : display all filedata fields,
 *  P_list_frag : display only fragments for each file,
 *  P_list_cluster : group files by base.
 * @param comparefile File to be compared with every listed ones.
 * @return PANA_code_error.
 */
int panalist( const char *pathname, int options, const char * comparefile)
{
	// local vars
	pds vpds;
	pds::filedata fd;

	// compare vars
	pts cpts;
	pds cpds;


	// process options
	//
	bool long_o = options & 1;
	bool ts_o = options & 2;
	bool rec_o = options & 8;

	// check compare file
	//
	if ( comparefile != NULL)
	{
		// compare file is regular file
		//
		if ( ! System::isFile( comparefile))
		{
			ERROR( PANA_debug_commands,
				gettext("panalist: compare file is not a regular file."))
			return 1;
		}
	
		// compare file must be readable
		//
		if ( ! System::isReadable( comparefile))
		{
			ERROR( PANA_debug_commands,
				gettext("panalist: compare file doesn't have read permission."))
			return 1;
		}

		// get directory name of compare file
		//	
		char dircomp[buffer_size];
		System::getDirName( comparefile, dircomp, buffer_size);
		char namecomp[buffer_size];
		System::getFileName( comparefile, namecomp, buffer_size);

		// check compare pds file
		//
		string pdsfilename = dircomp;
		pdsfilename += System::fileSystemSeparator;
		pdsfilename += PTDEF_PDS_FILENAME;
		
		// compare pds file is regular file
		//
		if ( ! System::isFile( pdsfilename.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panalist: pds file is not a regular file."))
			return 1;
		}
	
		// pds file must be readable
		//
		if ( ! System::isReadable( pdsfilename.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panalist: pds file doesn't have read permission."))
			return 1;
		}

		// open pds directory
		//
		if ( ! cpds.openDir( dircomp))
		{
			ERROR( PANA_debug_commands,
				gettext("panalist: error in opening pds directory."))
			return 1;
		}
		
		// check if compare file is PANA file
		//
		if ( ! cpds.hasFile( namecomp))
		{
			ERROR( PANA_debug_commands,
				gettext("panalist: compare file is not a PANASYNC file."))
			return 1;
		}

		// update file data
		//
		cpds.updateFile( namecomp);

		// get time-stamp
		//
		cpts = cpds.getFileData( namecomp).stamp;

	} // end 'if has compare file'
	
	
	// check if pathname is directory
	//
	if ( System::isDir( pathname))
	{
		// directory must be readable
		//
		if ( ! System::isReadable( pathname))
		{
			ERROR( PANA_debug_commands,
				gettext("panalist: directory doesn't have read permission."))
			return 1;
		}

		// check if pds file exists
		//
		string pdsfilename = pathname;
		pdsfilename += System::fileSystemSeparator;
		pdsfilename += PTDEF_PDS_FILENAME;
		
		if ( System::exists( pdsfilename.c_str() ))
		{
			// pds file is regular file
			//
			if ( ! System::isFile( pdsfilename.c_str()))
			{
				ERROR( PANA_debug_commands,
					gettext("panalist: pds file is not a regular file."))
				return 1;
			}
	
			// target pds file must be readable
			//
			if ( ! System::isReadable( pdsfilename.c_str()))
			{
				ERROR( PANA_debug_commands,
					gettext("panalist: pds file doesn't have read permission."))
				return 1;
			}
		} // end 'if target pds file exists'

		// open pds directory
		//
		if ( ! vpds.openDir( pathname))
		{
			ERROR( PANA_debug_commands,
				gettext("panalist: error in opening pds directory."))
			return 1;
		}

		// print directory path
		//
		cout << pathname << ":" << endl;

		// get files
		//
		vector<string> l = vpds.listFileNames();
		
		// for each pana-file
		//
		for ( vector<string>::iterator i = l.begin(); i!=l.end(); i++)
		{
			// update information
			//
			vpds.updateFile( i->c_str() );

			// get filedata for file
			//
			fd = vpds.getFileData( i->c_str());

			if ( comparefile != NULL )
			{
				cout << comparefile << " ";
				if (cpts.uncomparable( fd.stamp))
					cout << " ? ";
				else if (cpts.equivalent( fd.stamp))
					cout << "== ";
				else if (cpts.dominates( fd.stamp))
					cout << " > ";
				else if ( fd.stamp.dominates(cpts))
					cout << " < ";
				else
					cout << "|| ";
			}

			// print information for file
			//
			cout << i->c_str() 
					 << "\t";
			if ( long_o )
			{
				cout << "[";
				cout < fd.md5;
				cout << "] " << fd.stamp;
			}
			else if ( ts_o )
			{
				cout << fd.stamp;
			}
			cout << endl;
		} // end 'for each file'

		// if empty print '-'
		//
		if ( vpds.empty())
			cout << "---" << endl;
		cout << endl;

		// if recursive call for sub-dirs
		//
		if ( rec_o )
		{
			l = System::listDirs( pathname);
			for ( vector<string>::iterator i=l.begin(); i!=l.end(); i++)
			{
				// System::listDirs only returns the name of the sub-dir
				// not the complete or parcial path, so...
				// the path must be inserted before the sub-dir name
				//
				// create a path without last filesystemseparator
				char buffer[buffer_size];
				strncpy(buffer, pathname, buffer_size);
				System::removeLastSeparator( buffer);

				// prepend path (in buffer) to the sub-dir name
				//
				i->insert( (unsigned)0, (unsigned)1, System::fileSystemSeparator);
				i->insert( 0, buffer);

				panalist( i->c_str(), options, comparefile);
			}
		}

	}
	// else pathname is file
	else if ( System::isFile( pathname))
	{
		// check directory
		//
		char dirname[buffer_size];
		char filename[buffer_size];

		System::getDirName( pathname, dirname, buffer_size);
		System::getFileName( pathname, filename, buffer_size);

		// directory must be readable
		//
		if ( ! System::isReadable( dirname))
		{
			ERROR( PANA_debug_commands,
				gettext("panalist: file directory is not readable."))
			return 1;
		}
	
		// pds file
		//
		string pdsname = dirname;
		pdsname += PTDEF_PDS_FILENAME;
		
		if ( ! System::isFile( pdsname.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panalist: pds file is not a regular file."))
			return 1;
		}
	
		// pds file must be readable
		//
		if ( ! System::isReadable( pdsname.c_str()))
		{
			ERROR( PANA_debug_commands,
				gettext("panalist: pds file doesn't have read permission."))
			return 1;
		}
	
		// open pds directory
		//
		if ( ! vpds.openDir( dirname))
		{
			ERROR( PANA_debug_commands,
				gettext("panalist: error when opening pds directory."))
			return 1;
		}
		// if pathname is a PANA file
		if ( vpds.hasFile( filename))
		{
			// update file info
			//
			vpds.updateFile( filename);

			// get filedata for filename
			//
			fd = vpds.getFileData( filename);

			if ( comparefile != NULL )
			{
				cout << comparefile << " ";
				if (cpts.uncomparable( fd.stamp))
					cout << " ? ";
				else if (cpts.equivalent( fd.stamp))
					cout << "== ";
				else if (cpts.dominates( fd.stamp))
					cout << " > ";
				else if ( fd.stamp.dominates(cpts))
					cout << " < ";
				else
					cout << "|| ";
			}

			// print information for file
			//
			cout << pathname << "\t";
			if ( long_o )
			{
				cout << "[";
				cout < fd.md5;
				cout << "] " << fd.stamp;
			}
			else if ( ts_o )
			{
				cout << fd.stamp;
			}
			cout << endl;

		}
		else // not a PANA file
		{
			cout << "---" << endl << endl;
		}

	}
	else // don't know what it is!
	{
		ERROR( PANA_debug_commands,
			gettext("panalist: pathname is neither file or directory."))
		return 1;
	}

	return 0;
}
				

/** Make the translation between error codes and strings. 
 *
 * FIXME: not working yet!
 *
 * @param error_num The code of the panasync error.
 * @return The corresponding string if one applies.
 */
const char * panaerror( int error_num)
{
	switch (error_num)
	{
		case 0:
			return "panasync: success.";
		case 1:
			return "panasync: [internal error] not implemented yet!";
		case 2:
			return "";
		default:
			return "panasync: error unknown.";
	}

}

