/*
 * Panasync
 *
 * Copyright (C) 2001 Nuno Lopes
 * Copyright (C) 2000-2001 Carlos Baquero, Victor Fonte,
 * Paulo Sergio Almeida, Universidade do Minho
 *
 * panasync@gsd.di.uminho.pt - http://gsd.di.uminho.pt/panasync
 *
 * Departamento de Informatica, Universidade do Minho
 * Campus de Gualtar, 4710-057 Braga, Portugal
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 *
 */


#include "pts.h"


#ifdef	__panasync_debug_causal_history
int cch = 0;
#endif  //  __panasync_debug_causal_history

//      implementation of pts methods
//      ------------------------------------------------------------------------

pts::pts(){
}


pts::pts (string kind, string id, string counter)
{
    bs.kind    = kind;
    bs.id      = id;
    bs.counter = counter;
}

/** Creates a new identical time-stamp from another. */
pts::pts (const pts & source)
{
    bs = source.bs;
    ff = source.ff;
    sf = source.sf;
}

/** Creates a new identical time-stamp from a base. */
pts::pts(const base & b)
{
	bs.kind    = b.kind;
	bs.id      = b.id;
	bs.counter = b.counter;
}


pts::~pts ()
{
}

void pts::set_dirty ()
{
    ff = sf;
#ifdef	__panasync_debug_causal_history
    ch.insert (cch++);
#endif  //  __panasync_debug_causal_history
}

//
// definition of part dominates 
// ------------------------------------------------------------------------

bool dominates (const pts::part & p1, const pts::part & p2)
{
    return p1.size () > p2.size ()
      && equal (p2.begin (), p2.end (), p1.begin ());
}

struct fodominates:public binary_function < pts::part, pts::part, bool >
{
    bool operator () (const pts::part & p1, const pts::part & p2) const
    { return dominates (p1, p2); }
};

bool dominates (const pts::frag & f1, const pts::frag & f2)
{
    if (f1.size () == f2.size ()
	&& equal (f1.begin (), f1.end (), f2.begin ()))
	return false;
    for (pts::frag::iterator i = f2.begin (); i != f2.end (); ++i)
    {
	if (find_if (f1.begin (), f1.end (),
		     compose2 (logical_or < bool > (),
			       bind2nd (fodominates (), *i),
			       bind2nd (equal_to < pts::part > (),
					*i))) == f1.end ())
	    return false;
    }
    return true;
}

void pts::distribute (frag & ds, frag & fs1, frag & fs2, frag & fs3)
{
    if (ds.size () == 1)
    {
	part p1 = *ds.begin ();
	part p2 = *ds.begin ();

	p1.push_back (false);
	fs1.insert (p1);
	p2.push_back (true);
	fs2.insert (p2);
	fs3.erase (*ds.begin ());
    }
    else
    {
	for (frag::iterator i = ds.begin (); i != ds.end (); i++)
	{
	    (fs1.size () < fs2.size ()? fs1 : fs2).insert (*i);
	    fs3.erase (*i);
	}
    }
}

pts pts::fork ()
{
    pts npts (*this);

    if (sf.empty ())
    {
	sf.insert (part (1, false));
	npts.sf.insert (part (1, true));
    }
    else
    {
	frag fset = sf, fsetx;

	sf.clear ();
	npts.sf.clear ();

	for (frag::iterator k = ff.begin (); k != ff.end (); k++)
	{
	    // collect all sf elems that dominate a given ff elem
	    frag dset;

	    for (frag::iterator l = fset.begin (); l != fset.end (); l++)
		if (::dominates (*l, *k) || *l == *k)
		    dset.insert (*l);

	    distribute (dset, sf, npts.sf, fset);
	}

	if (ff.empty ())
	    distribute (fset, sf, npts.sf, fset);

	for (frag::iterator i = fset.begin (); i != fset.end (); i++)
	    (sf.size () < npts.sf.size ()? sf : npts.sf).insert (*i);
    }

    return npts;
}

ostream & operator << (ostream & out, const pts::part & part);
ostream & operator << (ostream & out, const pts::frag & frag);

struct is_sibling
{

    bool operator () (const pts::part & p1, const pts::part & p2)
    {
	assert (p1.size () != 0 && p2.size () != 0);

	return p1.size () == p2.size () &&
	  equal (p1.begin (), p1.end () - 1, p2.begin ());
    }
};

pts & pts::join (const pts & relative)	// bem,  completamente  bruta
{
    frag::iterator p;

    // join sets in second frag
    // note that duplicated pts::part objects are implicitly eliminated
    sf.insert (relative.sf.begin (), relative.sf.end ());

    // find & reduce siblings
//      cout << "join: before reduce " << *this;
    while ((p = adjacent_find (sf.begin (), sf.end (), is_sibling ())) !=
	   sf.end ())
    {
	part np = *p;
	part npn = *(++p);

//              cout << "reducing siblings: " << np << "+" << npn << "\n";
	sf.erase (np);
	np.pop_back ();
	sf.erase (npn);
	sf.insert (np);
    }
//      cout << "join: after reduce " << *this;

    // join first frag
    ff.insert (relative.ff.begin (), relative.ff.end ());

    // find & reduce siblings
//      cout << "join: before reduce " << *this;
    while ((p = adjacent_find (ff.begin (), ff.end (), is_sibling ())) !=
	   ff.end ())
    {
	part np = *p;
	part npn = *(++p);

//              cout << "reducing siblings: " << np << "+" << npn << "\n";
	ff.erase (np);
	np.pop_back ();
	ff.erase (npn);
	if (!np.empty ())
	    ff.insert (np);
    }
//      cout << "join: after reduce " << *this;

    //      find and reduce dominations in ff
    frag::reverse_iterator i, j;
    frag::iterator k, l;

    for (i = ff.rbegin (); i != ff.rend (); ++i)	// reduce
    {
	if ((j = find_if (i, ff.rend (), bind1st (fodominates (), *i))) !=
	    ff.rend ())
	    ff.erase (*j);
    }

    // find and reduce transform based dominations in ff
	 // this 3rd phase has been provved to be wrong so it is commented out
	 // since May 2001
/*
    frag t;
    part pt;
    bool b;

    for (k = ff.begin (); k != ff.end (); ++k)	// transform
    {
	pt = *k;
	b = pt.back ();
	pt.pop_back ();
	pt.push_back (!b);

	t.insert (pt);
    }

    for (i = t.rbegin (); i != t.rend (); ++i)	// reduce
    {
	if ((j = find_if (i, t.rend (), bind1st (fodominates (), *i))) !=
	    t.rend ())
	    t.erase (*j);
    }

    ff.clear ();
    for (k = t.begin (); k != t.end (); ++k)	// tranform back
    {
	pt = *k;
	b = pt.back ();
	pt.pop_back ();
	pt.push_back (!b);

	ff.insert (pt);
    }

	 */


    // remove eventual extra detail in ff by comparing with sf
    for (k = ff.begin (); k != ff.end (); ++k)
    {
	if ((l = find_if (sf.begin (), sf.end (),
			  compose2 (logical_or < bool > (),
				    bind2nd (fodominates (), *k),
				    bind2nd (equal_to < pts::part > (),
					     *k)))) == sf.end ())
	{			// found one that is not dominated
	    if (
		(l =
		 find_if (sf.begin (), sf.end (),
			  bind1st (fodominates (), *k))) != sf.end ())
	    {
		ff.erase (*k);
		ff.insert (*l);
	    }
	    else
		ff.erase (*k);
	}
    }

#ifdef __panadebug__causal_history
    ch.insert (relative.ch.begin (), relative.ch.end ());
#endif  // __panadebug__causal_history

	 return *this;
}

bool pts::dominates (const pts & relative) const
{
    bool r =::dominates (ff, relative.ff);

#ifdef __panadebug__causal_history
    assert (
	    (r == true && includes (ch.begin (), ch.end (),
				    relative.ch.begin (), relative.ch.end ()))
	    ||
	    (r == false &&
	     (!includes (ch.begin (), ch.end (),
			 relative.ch.begin (), relative.ch.end ()))
	     || equal (ch.begin (), ch.end (), relative.ch.begin ())));
#endif  // __panadebug__causal_history

		// NL inserted this,
		// please check if it is OK ?!
		return r;
}

bool pts::equivalent (const pts & relative) const
{
    return ff == relative.ff;
}

bool pts::uncomparable (const pts & relative) const
{
    return (bs.kind != relative.bs.kind ||
			   bs.id   != relative.bs.id   ||
			   bs.counter != relative.bs.counter);
}


template < class Stream > void pts::composite (Stream & stream)
{
    stream.simple(bs.kind).simple(bs.id).simple(bs.counter);
    stream.container(ff).container(sf);
}


#ifdef PHONY_F
// fake function
// important to force compilation of pts::composite
// is the compiler lazy?! compiling only on request or need...
// added by NL
void pts::phony(){

	// initialization of needed vars
	unsigned char fmb[PTBUFSIZE];
	mem_buffer xmb(fmb, sizeof(fmb));
	GIOP_format<mem_buffer> gfmb(xmb);

	// force compilation of pts:composite<input>
	obj_input<GIOP_format<mem_buffer> > is(gfmb);
	is.simple( *this);

	// force compilation of pts:composite<output>
	obj_output<GIOP_format<mem_buffer> > os(gfmb);
	os.simple( *this);
}
#endif  // PHONY_F

/** Get strig'alized base of instance.
 */
string pts::getBase()
{
	string ret;

	ret = bs.kind;
	ret += ".";
	ret += bs.id;
	ret += ".";
	ret += bs.counter;

	return ret;
}


#include <strstream.h>

bool operator < (const pts & p1, const pts & p2)
{
    bool r;
    strstream s1, s2;

    s1 << p1;
    s2 << p2;
//      if (strlen(s1.str()) < strlen(s2.str())) return true;
    r = lexicographical_compare (s1.str (), s1.str () + strlen (s1.str ()),
				 s2.str (), s2.str () + strlen (s2.str ()));
//      cout << "fox" << s1.str() << s2.str() << r << "|" << endl;
    return r;
//      return strcmp(s1.str(), s2.str()) < 0; //p2.dominates(p1);
}

// redefinition of part order operator
// ------------------------------------------------------------------------

bool operator < (const pts::part & p1, const pts::part & p2)
{
//      cout << "fox!" << p1.size() << " " << p2.size() << "\n!" ;
    if (p1.size () == p2.size ())
	return
	  lexicographical_compare (p1.begin (), p1.end (), p2.begin (),
				   p2.end ());
    return p1.size () < p2.size ();
}

//      iostream operators
//      ------------------------------------------------------------------------

struct printpartbit:public unary_function < bool, void >
{ printpartbit (ostream & out):os (out)
    {
    } void operator () (bool x)
    { os << x ? '1' : '0';
    } ostream & os;
};

ostream & operator << (ostream & out, const pts::part & part)
{

    if (!part.empty ())
    {
	for_each (part.begin (), part.end () - 1, printpartbit (out));
	out << part.back ();
    }
    return out;
}

ostream & operator << (ostream & out, const pts::frag & frag)
{

    if (!frag.empty ())
    {
	copy (frag.begin (), --frag.end (),
	      ostream_iterator < pts::part > (out, "+"));
	out << *--frag.end ();
    }

/*
	vector<pts::part> cepos;
	copy(frag.begin(), frag.end(), back_inserter(cepos));

	if (!cepos.empty())
	{
		copy(cepos.begin(), cepos.end() - 1,
			ostream_iterator<pts::part>(out, "+"));
		out << cepos.back();
	}
*/
    return out;
}

ostream & operator << (ostream & out, const pts & pts)
{
		// base section
#ifdef PTS_FULL_OUTPUT
		out << "{" << pts.bs.kind
				<< '.' << pts.bs.id
				<< '.' << pts.bs.counter
			 	<< ' ';
#endif  // PTS_FULL_OUTPUT

		// fragments section
    out << "[ " << pts.ff << " : " << pts.sf << " ]";

#ifdef PTS_FULL_OUTPUT
		out << "}";
#endif  // PTS_FULL_OUTPUT

		return out;
}


